# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import abc
import json
import logging
import requests
import time
import typing


from base64 import b64decode
from dataclasses import dataclass
from datetime import datetime, timedelta
from collections.abc import Mapping
from decimal import Decimal
from solana.blockhash import Blockhash, BlockhashCache
from solana.keypair import Keypair
from solana.publickey import PublicKey
from solana.rpc.api import Client
from solana.rpc.commitment import Commitment, Processed, Finalized
from solana.rpc.providers.http import HTTPProvider
from solana.rpc.types import (
    DataSliceOpts,
    MemcmpOpts,
    RPCMethod,
    RPCResponse,
    TokenAccountOpts,
    TxOpts,
)
from solana.transaction import Transaction

from .constants import SOL_DECIMAL_DIVISOR
from .datetimes import local_now
from .instructionreporter import InstructionReporter
from .logmessages import expand_log_messages
from .text import indent_collection_as_str


_STUB_TRANSACTION_SIGNATURE: str = "stub-for-already-submitted-transaction-signature"


# # 🥭 CompoundException class
#
# A `CompoundException` exception can hold all exceptions that were raised when trying to read or
# write to Solana.
#
# The `Client` can rotate through multiple providers, switching when certain `Exception`s are thrown.
# This exception allows all those exceptions to be gathered and inspected should all providers fail.
#
class CompoundException(Exception):
    def __init__(self, name: str, all_exceptions: typing.Sequence[Exception]) -> None:
        super().__init__(f"[{name}] Multiple errors captured for event")
        self.name: str = name
        self.all_exceptions: typing.Sequence[Exception] = all_exceptions

    def __str__(self) -> str:
        if len(self.all_exceptions) == 1:
            return str(self.all_exceptions[0])

        details: str = indent_collection_as_str(self.all_exceptions)
        return f"""« CompoundException with {len(self.all_exceptions)} inner exceptions:
    {details}
»"""


# # 🥭 ClientException class
#
# A `ClientException` exception base class that allows trapping and handling rate limiting
# independent of other error handling.
#
class ClientException(Exception):
    def __init__(self, message: str, name: str, cluster_rpc_url: str) -> None:
        super().__init__(message)
        self.message: str = message
        self.name: str = name
        self.cluster_rpc_url: str = cluster_rpc_url

    def __str__(self) -> str:
        return f"« {type(self)} '{self.message}' from '{self.name}' on {self.cluster_rpc_url} »"

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 RateLimitException class
#
# A `RateLimitException` exception base class that allows trapping and handling rate limiting
# independent of other error handling.
#
class RateLimitException(ClientException):
    pass


# # 🥭 TooMuchBandwidthRateLimitException class
#
# A `TooMuchBandwidthRateLimitException` exception that specialises the `RateLimitException`
# for when too much bandwidth has been consumed.
#
class TooMuchBandwidthRateLimitException(RateLimitException):
    pass


# # 🥭 TooManyRequestsRateLimitException class
#
# A `TooManyRequestsRateLimitException` exception that specialises the `RateLimitException`
# for when too many requests have been sent in a short time.
#
class TooManyRequestsRateLimitException(RateLimitException):
    pass


# # 🥭 BlockhashNotFoundException class
#
# A `BlockhashNotFoundException` exception allows trapping and handling exceptions when a blockhash is sent that
# the node doesn't understand. This can happen when the blockhash is too old (and the node no longer
# considers it 'recent') or when it's too new (and hasn't yet made it to the node that is responding).
#
class BlockhashNotFoundException(ClientException):
    def __init__(
        self,
        name: str,
        cluster_rpc_url: str,
        blockhash: typing.Optional[Blockhash] = None,
    ) -> None:
        message: str = f"Blockhash '{blockhash}' not found on {cluster_rpc_url}."
        super().__init__(message, name, cluster_rpc_url)
        self.blockhash: typing.Optional[Blockhash] = blockhash

    def __str__(self) -> str:
        return f"« BlockhashNotFoundException '{self.name}' [{self.blockhash}] on {self.cluster_rpc_url} »"


# # 🥭 NodeIsBehindException class
#
# A `NodeIsBehindException` exception allows trapping and handling exceptions when a node is behind by too
# many slots.
#
class NodeIsBehindException(ClientException):
    def __init__(self, name: str, cluster_rpc_url: str, slots_behind: int) -> None:
        message: str = f"Node is behind by {slots_behind} slots."
        super().__init__(message, name, cluster_rpc_url)
        self.slots_behind: int = slots_behind

    def __str__(self) -> str:
        return f"« NodeIsBehindException '{self.name}' [behind by {self.slots_behind} slots] on {self.cluster_rpc_url} »"


# # 🥭 TransactionAlreadyProcessedException class
#
# A `TransactionAlreadyProcessedException` exception indicates that a transaction with that signature and
# recent blockhash has already been processed.
#
# This can be a normal consequence of the stale slot checking/resubmission (which will mean it is ignored)
# or it can indicate an actual problem.
#
class TransactionAlreadyProcessedException(RateLimitException):
    pass


# # 🥭 StaleSlotException class
#
# A `StaleSlotException` exception allows trapping and handling exceptions when data is received from
#
class StaleSlotException(ClientException):
    def __init__(
        self,
        name: str,
        cluster_rpc_url: str,
        latest_seen_slot: int,
        just_returned_slot: int,
    ) -> None:
        message: str = f"Stale slot received - received data from slot {just_returned_slot} having previously seen slot {latest_seen_slot}."
        super().__init__(message, name, cluster_rpc_url)
        self.latest_seen_slot: int = latest_seen_slot
        self.just_returned_slot: int = just_returned_slot

    def __str__(self) -> str:
        return f"« StaleSlotException '{self.name}' [received data from slot {self.just_returned_slot} having previously seen slot {self.latest_seen_slot}] on {self.cluster_rpc_url} »"


# # 🥭 FailedToFetchBlockhashException class
#
# A `FailedToFetchBlockhashException` exception allows trapping and handling exceptions when we fail
# to fetch a recent or distinct blockhash.
#
class FailedToFetchBlockhashException(ClientException):
    def __init__(
        self,
        message: str,
        name: str,
        cluster_rpc_url: str,
        pauses: typing.Sequence[float],
    ) -> None:
        super().__init__(message, name, cluster_rpc_url)
        self.pauses: typing.Sequence[float] = pauses

    def __str__(self) -> str:
        if len(self.pauses) == 0:
            return f"« FailedToFetchBlockhashException '{self.name}' Failed to get recent blockhash on {self.cluster_rpc_url} »"

        pauses_text = ",".join(f"{pause}" for pause in self.pauses[:-1])
        return f"« FailedToFetchBlockhashException '{self.name}' Failed to get a fresh, recent blockhash after {len(self.pauses)} attempts - paused {pauses_text} seconds between attempts on {self.cluster_rpc_url} »"


# # 🥭 TransactionException class
#
# A `TransactionException` exception that can provide additional error data, or at least better output
# of problems at the right place.
#
class TransactionException(ClientException):
    def __init__(
        self,
        transaction: typing.Optional[Transaction],
        message: str,
        code: int,
        name: str,
        cluster_rpc_url: str,
        rpc_method: str,
        request_text: str,
        response_text: str,
        accounts: typing.Union[str, typing.List[str], None],
        errors: typing.Union[str, typing.List[str], None],
        logs: typing.Union[str, typing.List[str], None],
        instruction_reporter: InstructionReporter = InstructionReporter(),
    ) -> None:
        super().__init__(message, name, cluster_rpc_url)
        self.transaction: typing.Optional[Transaction] = transaction
        self.code: int = code
        self.rpc_method: str = rpc_method
        self.request_text: str = request_text
        self.response_text: str = response_text

        def _ensure_list(
            item: typing.Union[str, typing.List[str], None]
        ) -> typing.List[str]:
            if item is None:
                return []
            if isinstance(item, str):
                return [item]
            if isinstance(item, list):
                return item
            return [f"{item}"]

        self.accounts: typing.Sequence[str] = _ensure_list(accounts)
        self.errors: typing.Sequence[str] = _ensure_list(errors)
        self.logs: typing.Sequence[str] = expand_log_messages(_ensure_list(logs))
        self.instruction_reporter: InstructionReporter = instruction_reporter

    def __str__(self) -> str:
        try:
            transaction_details = ""
            if self.transaction is not None:
                instruction_details = "\n".join(
                    list(
                        map(
                            self.instruction_reporter.report,
                            self.transaction.instructions,
                        )
                    )
                )
                transaction_details = (
                    "\n    Instructions:\n        "
                    + instruction_details.replace("\n", "\n        ")
                )
            accounts = "No Accounts"
            if len(self.accounts) > 0:
                accounts = "\n        ".join(
                    [f"{item}".replace("\n", "\n        ") for item in self.accounts]
                )
            errors = "No Errors"
            if len(self.errors) > 0:
                errors = "\n        ".join(
                    [f"{item}".replace("\n", "\n        ") for item in self.errors]
                )
            logs = "No Logs"
            if len(self.logs) > 0:
                logs = "\n        ".join(
                    [f"{item}".replace("\n", "\n        ") for item in self.logs]
                )
            return f"""« TransactionException in '{self.name}' [{self.rpc_method}]: {self.code}:: {self.message}{transaction_details}
    Accounts:
        {accounts}
    Errors:
        {errors}
    Logs:
        {logs}
»"""
        except Exception as exception:
            return f"TransactionException printing failed with: {exception}"

    def __repr__(self) -> str:
        return f"{self}"


UnspecifiedCommitment = Commitment("unspecified")
UnspecifiedEncoding = "unspecified"


# # 🥭 AbstractSlotHolder class
#
# A `SlotHolder` shares the latest slot across multiple `RPCCaller`s.
#
class AbstractSlotHolder(metaclass=abc.ABCMeta):
    def __init__(self) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)

    @abc.abstractproperty
    def latest_slot(self) -> int:
        raise NotImplementedError(
            "AbstractSlotHolder.latest_slot() is not implemented on the base type."
        )

    def require_data_from_fresh_slot(
        self, latest_slot: typing.Optional[int] = None
    ) -> None:
        raise NotImplementedError(
            "AbstractSlotHolder.require_data_from_fresh_slot() is not implemented on the base type."
        )

    def is_acceptable(self, slot_to_check: int) -> bool:
        raise NotImplementedError(
            "AbstractSlotHolder.is_acceptable() is not implemented on the base type."
        )


# # 🥭 NullSlotHolder class
#
# A `NullSlotHolder` takes no slot-checking action.
#
class NullSlotHolder(AbstractSlotHolder):
    def __init__(self) -> None:
        super().__init__()

    @property
    def latest_slot(self) -> int:
        return 0

    def require_data_from_fresh_slot(
        self, latest_slot: typing.Optional[int] = None
    ) -> None:
        pass

    def is_acceptable(self, slot_to_check: int) -> bool:
        return True


# # 🥭 CheckingSlotHolder class
#
# A `CheckingSlotHolder` allows checking if a given slot is newer or equal to the last slot seen.
#
class CheckingSlotHolder(AbstractSlotHolder):
    def __init__(self) -> None:
        super().__init__()
        self.__latest_slot: int = 0

    @property
    def latest_slot(self) -> int:
        return self.__latest_slot

    def require_data_from_fresh_slot(
        self, latest_slot: typing.Optional[int] = None
    ) -> None:
        latest: int = latest_slot or self.latest_slot
        if latest >= self.latest_slot:
            self.__latest_slot = latest + 1
            self._logger.debug(
                f"Requiring data from slot {self.latest_slot} onwards now."
            )

    def is_acceptable(self, slot_to_check: int) -> bool:
        if slot_to_check < self.__latest_slot:
            return False

        if slot_to_check > self.__latest_slot:
            self.__latest_slot = slot_to_check
            self._logger.debug(
                f"Only accepting data from slot {self.latest_slot} onwards now."
            )
        return True


class TransactionMonitor(metaclass=abc.ABCMeta):
    def __init__(
        self,
        commitment: Commitment = Finalized,
        transaction_timeout: float = 90.0,
        slot_holder: AbstractSlotHolder = NullSlotHolder(),
    ) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.commitment: Commitment = commitment
        self.transaction_timeout: float = transaction_timeout

        self.slot_holder: AbstractSlotHolder = slot_holder

    @abc.abstractmethod
    def monitor(self, signature: str) -> None:
        raise NotImplementedError(
            "TransactionMonitor.monitor() is not implemented on the base type."
        )

    def dispose(self) -> None:
        pass


class NullTransactionMonitor(TransactionMonitor):
    def __init__(
        self,
        commitment: Commitment = Finalized,
        transaction_timeout: float = 90.0,
        slot_holder: AbstractSlotHolder = NullSlotHolder(),
    ) -> None:
        super().__init__(
            commitment=commitment,
            transaction_timeout=transaction_timeout,
            slot_holder=slot_holder,
        )

    def monitor(self, signature: str) -> None:
        pass


# # 🥭 RPCCaller class
#
# A `RPCCaller` extends the HTTPProvider with better error handling.
#
class RPCCaller(HTTPProvider):
    def __init__(
        self,
        name: str,
        cluster_rpc_url: str,
        cluster_ws_url: str,
        http_request_timeout: float,
        stale_data_pauses_before_retry: typing.Sequence[float],
        slot_holder: AbstractSlotHolder,
        instruction_reporter: InstructionReporter,
    ):
        super().__init__(cluster_rpc_url)
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.name: str = name
        self.cluster_rpc_url: str = cluster_rpc_url
        self.cluster_ws_url: str = cluster_ws_url
        self.http_request_timeout: float = http_request_timeout
        self.stale_data_pauses_before_retry: typing.Sequence[
            float
        ] = stale_data_pauses_before_retry
        self.slot_holder: AbstractSlotHolder = slot_holder
        self.instruction_reporter: InstructionReporter = instruction_reporter

    def require_data_from_fresh_slot(
        self, latest_slot: typing.Optional[int] = None
    ) -> None:
        self.slot_holder.require_data_from_fresh_slot(latest_slot)

    def make_request(self, method: RPCMethod, *params: typing.Any) -> RPCResponse:
        # No pauses specified means this funcitonality is turned off.
        if len(self.stale_data_pauses_before_retry) == 0:
            return self.__make_request(method, *params)

        at_least_one_submission: bool = False
        last_stale_slot_exception: StaleSlotException
        for pause in [*self.stale_data_pauses_before_retry, 0]:
            try:
                return self.__make_request(method, *params)
            except TransactionAlreadyProcessedException as transaction_already_processed_exception:
                if not at_least_one_submission:
                    raise transaction_already_processed_exception

                # The transaction has already been processed (even though we previously got a stale slot
                # response). So, yay? It's good that it was successfully handled (and presumably processed)
                # and it means we can safely ignore this exception, but it does leave us with no proper
                # response to return.
                #
                # Return fake data in the expected structure for now, and try to figure out a better way.
                return {
                    "jsonrpc": "2.0",
                    "id": 0,
                    "result": _STUB_TRANSACTION_SIGNATURE,
                }
            except StaleSlotException as exception:
                last_stale_slot_exception = exception
                self._logger.debug(
                    f"Will retry after pause of {pause} seconds after getting stale slot: {exception}"
                )
                time.sleep(pause)
            at_least_one_submission = True

        # They've all failed.
        raise last_stale_slot_exception

    def __make_request(self, method: RPCMethod, *params: typing.Any) -> RPCResponse:
        # This is the entire method in HTTPProvider that we're overriding here:
        #
        # """Make an HTTP request to an http rpc endpoint."""
        # request_kwargs = self._before_request(method=method, params=params, is_async=False)
        # raw_response = requests.post(**request_kwargs)
        # return self._after_request(raw_response=raw_response, method=method)

        request_kwargs = self._before_request(
            method=method, params=params, is_async=False
        )
        http_post_timeout: typing.Union[float, None] = (
            self.http_request_timeout if self.http_request_timeout >= 0 else None
        )
        raw_response = requests.post(**request_kwargs, timeout=http_post_timeout)

        # Some custom exceptions specifically for rate-limiting. This allows calling code to handle this
        # specific case if they so choose.
        #
        # "You will see HTTP respose codes 429 for too many requests or 413 for too much bandwidth."
        if raw_response.status_code == 413:
            raise TooMuchBandwidthRateLimitException(
                f"Rate limited (too much bandwidth) calling method '{method}' on {self.cluster_rpc_url}",
                self.name,
                self.cluster_rpc_url,
            )
        elif raw_response.status_code == 429:
            raise TooManyRequestsRateLimitException(
                f"Rate limited (too many requests) calling method '{method}' on {self.cluster_rpc_url}",
                self.name,
                self.cluster_rpc_url,
            )

        # Not a rate-limit problem, but maybe there was some other error?
        raw_response.raise_for_status()

        # All seems OK, but maybe the server returned an error? If so, try to pass on as much
        # information as we can.
        response_text: str = raw_response.text
        response: typing.Dict[str, typing.Any] = json.loads(response_text)

        # Did we get sufficiently up-to-date information? It must be from the last slot we saw or a
        # newer slot.
        #
        # Only do this check if we're using a commitment level of 'processed'.
        if (
            isinstance(params, Mapping)
            and len(params) > 1
            and "commitment" in params[1]
            and params[1]["commitment"] == Processed
        ):
            if (
                "result" in response
                and isinstance(response["result"], Mapping)
                and "context" in response["result"]
                and isinstance(response["result"]["context"], Mapping)
                and "slot" in response["result"]["context"]
            ):
                slot: int = response["result"]["context"]["slot"]
                if not self.slot_holder.is_acceptable(slot):
                    self._logger.warning(
                        f"Result is from stale slot: {slot} - latest slot is: {self.slot_holder.latest_slot}"
                    )
                    raise StaleSlotException(
                        self.name,
                        self.cluster_rpc_url,
                        self.slot_holder.latest_slot,
                        slot,
                    )

        if "error" in response:
            if response["error"] is str:
                message: str = typing.cast(str, response["error"])
                raise ClientException(
                    f"Transaction failed: '{message}'", self.name, self.cluster_rpc_url
                )
            else:
                error = response["error"]
                error_message: str = (
                    error["message"] if "message" in error else "No message"
                )
                error_data: typing.Dict[str, typing.Any] = (
                    error["data"] if "data" in error else {}
                )
                error_accounts = (
                    error_data["accounts"]
                    if "accounts" in error_data
                    else "No accounts"
                )
                error_code: int = error["code"] if "code" in error else -1
                error_err = (
                    error_data["err"]
                    if "err" in error_data
                    else "No error text returned"
                )
                error_logs = error_data["logs"] if "logs" in error_data else "No logs"
                parameters = json.dumps(
                    {"jsonrpc": "2.0", "method": method, "params": params}
                )

                transaction: typing.Optional[Transaction] = None
                blockhash: typing.Optional[Blockhash] = None
                if method == "sendTransaction":
                    transaction = Transaction.deserialize(b64decode(params[0]))
                    blockhash = transaction.recent_blockhash

                if error_code == -32005:
                    slots_behind: int = (
                        error["data"]["numSlotsBehind"]
                        if "numSlotsBehind" in error["data"]
                        else -1
                    )
                    raise NodeIsBehindException(
                        self.name, self.cluster_rpc_url, slots_behind
                    )

                if error_err == "BlockhashNotFound":
                    raise BlockhashNotFoundException(
                        self.name, self.cluster_rpc_url, blockhash
                    )

                if error_err == "AlreadyProcessed":
                    raise TransactionAlreadyProcessedException(
                        error_message, self.name, self.cluster_rpc_url
                    )

                exception_message: str = f"Transaction failed with: '{error_message}'"
                raise TransactionException(
                    transaction,
                    exception_message,
                    error_code,
                    self.name,
                    self.cluster_rpc_url,
                    method,
                    parameters,
                    response_text,
                    error_accounts,
                    error_err,
                    error_logs,
                    self.instruction_reporter,
                )

        if method == "getRecentBlockhash":
            if (
                "result" in response
                and "value" in response["result"]
                and "blockhash" in response["result"]["value"]
                and "context" in response["result"]
                and "slot" in response["result"]["context"]
            ):
                fresh_blockhash = Blockhash(response["result"]["value"]["blockhash"])
                fresh_blockhash_slot = Blockhash(response["result"]["context"]["slot"])
                self._logger.debug(
                    f"Recent blockhash [slot: {fresh_blockhash_slot}]: {fresh_blockhash}"
                )

        # The call succeeded.
        return typing.cast(RPCResponse, response)

    def __str__(self) -> str:
        return f"« RPCCaller [{self.cluster_rpc_url}] »"

    def __repr__(self) -> str:
        return f"{self}"


# # 🥭 CompoundRPCCaller class
#
# A `CompoundRPCCaller` will try multiple providers until it succeeds (or the all fail). Should only trap
# and switch provider on exceptions that show that provider is no longer at the tip of the chain.
#
class CompoundRPCCaller(HTTPProvider):
    def __init__(self, name: str, providers: typing.Sequence[RPCCaller]):
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.__providers: typing.Sequence[RPCCaller] = providers
        self.name: str = name
        self.on_provider_change: typing.Callable[[], None] = lambda: None

        # Some libraries (like the SPL Token Library) depend on `endpoint_uri` existing on
        # a `Provider`, and static typing complains if we use a property instead of a
        # variable. We add it here and keep it up to date when we switch inner providers.
        self.endpoint_uri = providers[0].endpoint_uri

    @property
    def current(self) -> RPCCaller:
        return self.__providers[0]

    @property
    def all_providers(self) -> typing.Sequence[RPCCaller]:
        return self.__providers

    def shift_to_next_provider(self) -> None:
        # This is called when the current provider is raising errors, when the next provider might not.
        # Typical RPC host errors are trapped and managed via make_request(), but some errors can't be
        # handled properly there. For example, BlockhashNotFound exceptions can be trapped there, but
        # the right answer is to switch to the next provider AND THEN fetch a fresh blockhash and retry
        # the transaction. That's not possible to do atomically (without a lot of nasty, fragile work) so
        # it's better to handle it at the higher level. That's what this method allows - the higher level
        # can call this to switch to the next provider, and it can then fetch the fresh blockhash and
        # resubmit the transaction.
        if len(self.__providers) > 1:
            self.__providers = [*self.__providers[1:], *self.__providers[:1]]
            self.endpoint_uri = self.__providers[0].endpoint_uri
            self.on_provider_change()
        self._logger.debug(f"Told to shift provider - now using: {self.__providers[0]}")

    def make_request(self, method: RPCMethod, *params: typing.Any) -> RPCResponse:
        all_exceptions: typing.List[Exception] = []
        for provider in self.__providers:
            try:
                result = provider.make_request(method, *params)
                successful_index: int = self.__providers.index(provider)
                if successful_index != 0:
                    # Rebase the providers' list so we continue to use this successful one (until it fails)
                    self.__providers = [
                        *self.__providers[successful_index:],
                        *self.__providers[:successful_index],
                    ]
                    self.endpoint_uri = self.__providers[0].endpoint_uri
                    self.on_provider_change()
                    self._logger.debug(
                        f"Shifted provider - now using: {self.__providers[0]}"
                    )
                return result
            except (
                requests.exceptions.HTTPError,
                requests.exceptions.ConnectionError,
                requests.exceptions.Timeout,
                RateLimitException,
                NodeIsBehindException,
                StaleSlotException,
                FailedToFetchBlockhashException,
            ) as exception:
                all_exceptions += [exception]
                self._logger.info(
                    f"Moving to next provider - {provider} gave {exception}"
                )

        if len(all_exceptions) == 1:
            raise all_exceptions[0]

        raise CompoundException(self.name, all_exceptions)

    def is_connected(self) -> bool:
        # All we need for this to be true is for one of our providers to be connected.
        for provider in self.__providers:
            if provider.is_connected():
                return True
        return False

    def __str__(self) -> str:
        return f"« CompoundRPCCaller with {len(self.__providers)} providers - current head is: {self.__providers[0]} »"

    def __repr__(self) -> str:
        return f"{self}"


# # Data transfer class for cluster url
#
@dataclass
class ClusterUrlData:
    rpc: str  # http/https connection to contact RPC nodes
    ws: str = ""  # ws/wss connection to register to RPC websocket

    def __post_init__(self) -> None:
        if not self.ws:
            self.ws = self.rpc.replace("https:", "wss:", 1)
            self.ws = self.ws.replace("http:", "ws:", 1)

    def __str__(self) -> str:
        return f"« ClusterUrlData RPC: {self.rpc} WS: {self.ws} »"

    def __repr__(self) -> str:
        return f"{self}"


class BetterClient:
    def __init__(
        self,
        client: Client,
        name: str,
        cluster_name: str,
        commitment: Commitment,
        skip_preflight: bool,
        tpu_retransmissions: int,
        encoding: str,
        blockhash_cache_duration: int,
        rpc_caller: CompoundRPCCaller,
        transaction_monitor: TransactionMonitor = NullTransactionMonitor(),
    ) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.compatible_client: Client = client
        self.name: str = name
        self.cluster_name: str = cluster_name
        self.commitment: Commitment = commitment
        self.skip_preflight: bool = skip_preflight
        self.tpu_retransmissions: int = tpu_retransmissions
        self.encoding: str = encoding
        self.blockhash_cache_duration: int = blockhash_cache_duration
        self.rpc_caller: CompoundRPCCaller = rpc_caller
        self.transaction_monitor: TransactionMonitor = transaction_monitor

    @staticmethod
    def from_configuration(
        name: str,
        cluster_name: str,
        cluster_urls: typing.Sequence[ClusterUrlData],
        commitment: Commitment,
        skip_preflight: bool,
        tpu_retransmissions: int,
        encoding: str,
        blockhash_cache_duration: int,
        http_request_timeout: float,
        stale_data_pauses_before_retry: typing.Sequence[float],
        instruction_reporter: InstructionReporter,
        transaction_monitor: TransactionMonitor = NullTransactionMonitor(),
    ) -> "BetterClient":
        rpc_callers: typing.List[RPCCaller] = []
        for cluster_url in cluster_urls:
            rpc_caller: RPCCaller = RPCCaller(
                name,
                cluster_url.rpc,
                cluster_url.ws,
                http_request_timeout,
                stale_data_pauses_before_retry,
                transaction_monitor.slot_holder,
                instruction_reporter,
            )
            rpc_callers += [rpc_caller]

        provider: CompoundRPCCaller = CompoundRPCCaller(name, rpc_callers)
        blockhash_cache: typing.Union[BlockhashCache, bool] = False
        if blockhash_cache_duration > 0:
            blockhash_cache = BlockhashCache(blockhash_cache_duration)
        client: Client = Client(
            endpoint=cluster_urls[0].rpc,
            commitment=commitment,
            blockhash_cache=blockhash_cache,
        )
        client._provider = provider

        def __on_provider_change() -> None:
            if client.blockhash_cache:
                # Clear out the blockhash cache on retrying
                logging.debug("Replacing client blockhash cache.")
                client.blockhash_cache = BlockhashCache(blockhash_cache_duration)
                blockhash_resp = client.get_recent_blockhash(Finalized)
                client._process_blockhash_resp(blockhash_resp, used_immediately=False)

        provider.on_provider_change = __on_provider_change

        return BetterClient(
            client,
            name,
            cluster_name,
            commitment,
            skip_preflight,
            tpu_retransmissions,
            encoding,
            blockhash_cache_duration,
            provider,
            transaction_monitor,
        )

    @property
    def cluster_rpc_url(self) -> str:
        return self.rpc_caller.current.cluster_rpc_url

    @property
    def cluster_rpc_urls(self) -> typing.Sequence[str]:
        return [
            rpc_caller.cluster_rpc_url for rpc_caller in self.rpc_caller.all_providers
        ]

    @property
    def cluster_ws_url(self) -> str:
        return self.rpc_caller.current.cluster_ws_url

    @property
    def cluster_ws_urls(self) -> typing.Sequence[str]:
        return [
            rpc_caller.cluster_ws_url for rpc_caller in self.rpc_caller.all_providers
        ]

    @property
    def cluster_urls(self) -> typing.Sequence[ClusterUrlData]:
        return [
            ClusterUrlData(rpc=rpc_caller.cluster_rpc_url, ws=rpc_caller.cluster_ws_url)
            for rpc_caller in self.rpc_caller.all_providers
        ]

    @property
    def instruction_reporter(self) -> InstructionReporter:
        return self.rpc_caller.current.instruction_reporter

    @property
    def stale_data_pauses_before_retry(self) -> typing.Sequence[float]:
        return self.rpc_caller.current.stale_data_pauses_before_retry

    def dispose(self) -> None:
        self.transaction_monitor.dispose()

    def require_data_from_fresh_slot(self) -> None:
        self.rpc_caller.current.require_data_from_fresh_slot()

    def get_balance(
        self,
        pubkey: typing.Union[PublicKey, str],
        commitment: Commitment = UnspecifiedCommitment,
    ) -> Decimal:
        resolved_commitment, _ = self.__resolve_defaults(commitment)
        response = self.compatible_client.get_balance(pubkey, resolved_commitment)
        value = Decimal(response["result"]["value"])
        return value / SOL_DECIMAL_DIVISOR

    def get_account_info(
        self,
        pubkey: typing.Union[PublicKey, str],
        commitment: Commitment = UnspecifiedCommitment,
        encoding: str = UnspecifiedEncoding,
        data_slice: typing.Optional[DataSliceOpts] = None,
    ) -> typing.Any:
        resolved_commitment, resolved_encoding = self.__resolve_defaults(
            commitment, encoding
        )
        response = self.compatible_client.get_account_info(
            pubkey, resolved_commitment, resolved_encoding, data_slice
        )
        return response["result"]

    def get_confirmed_signatures_for_address2(
        self,
        account: typing.Union[str, Keypair, PublicKey],
        before: typing.Optional[str] = None,
        until: typing.Optional[str] = None,
        limit: typing.Optional[int] = None,
    ) -> typing.Sequence[str]:
        response = self.compatible_client.get_confirmed_signature_for_address2(
            account, before, until, limit
        )
        return [result["signature"] for result in response["result"]]

    def get_confirmed_transaction(
        self, signature: str, encoding: str = "json"
    ) -> typing.Any:
        _, resolved_encoding = self.__resolve_defaults(None, encoding)
        response = self.compatible_client.get_confirmed_transaction(
            signature, resolved_encoding
        )
        return response["result"]

    def get_minimum_balance_for_rent_exemption(
        self, size: int, commitment: Commitment = UnspecifiedCommitment
    ) -> int:
        resolved_commitment, _ = self.__resolve_defaults(commitment)
        response = self.compatible_client.get_minimum_balance_for_rent_exemption(
            size, resolved_commitment
        )
        return int(response["result"])

    def get_program_accounts(
        self,
        pubkey: typing.Union[str, PublicKey],
        commitment: Commitment = UnspecifiedCommitment,
        encoding: typing.Optional[str] = UnspecifiedEncoding,
        data_slice: typing.Optional[DataSliceOpts] = None,
        data_size: typing.Optional[int] = None,
        memcmp_opts: typing.Optional[typing.List[MemcmpOpts]] = None,
    ) -> typing.Any:
        resolved_commitment, resolved_encoding = self.__resolve_defaults(
            commitment, encoding
        )
        response = self.compatible_client.get_program_accounts(
            pubkey,
            resolved_commitment,
            resolved_encoding,
            data_slice,
            data_size,
            memcmp_opts,
        )

        return response["result"]

    def get_recent_blockhash(
        self, commitment: Commitment = UnspecifiedCommitment
    ) -> Blockhash:
        resolved_commitment, _ = self.__resolve_defaults(commitment)
        response = self.compatible_client.get_recent_blockhash(resolved_commitment)
        return Blockhash(response["result"]["value"]["blockhash"])

    def get_token_account_balance(
        self,
        pubkey: typing.Union[str, PublicKey],
        commitment: Commitment = UnspecifiedCommitment,
    ) -> Decimal:
        resolved_commitment, _ = self.__resolve_defaults(commitment)
        response = self.compatible_client.get_token_account_balance(
            pubkey, resolved_commitment
        )
        value = Decimal(response["result"]["value"]["amount"])
        decimal_places = response["result"]["value"]["decimals"]
        divisor = Decimal(10**decimal_places)
        return value / divisor

    def get_token_accounts_by_owner(
        self,
        owner: PublicKey,
        token_account_options: TokenAccountOpts,
        commitment: Commitment = UnspecifiedCommitment,
    ) -> typing.Any:
        resolved_commitment, _ = self.__resolve_defaults(commitment)
        response = self.compatible_client.get_token_accounts_by_owner(
            owner, token_account_options, resolved_commitment
        )
        return response["result"]["value"]

    def get_multiple_accounts(
        self,
        pubkeys: typing.List[typing.Union[PublicKey, str]],
        commitment: Commitment = UnspecifiedCommitment,
        encoding: str = UnspecifiedEncoding,
        data_slice: typing.Optional[DataSliceOpts] = None,
    ) -> typing.Any:
        resolved_commitment, resolved_encoding = self.__resolve_defaults(
            commitment, encoding
        )
        response = self.compatible_client.get_multiple_accounts(
            pubkeys, resolved_commitment, resolved_encoding, data_slice
        )
        return response["result"]["value"]

    def send_transaction(
        self,
        transaction: Transaction,
        *signers: Keypair,
        opts: TxOpts = TxOpts(preflight_commitment=UnspecifiedCommitment),
        recent_blockhash: typing.Optional[Blockhash] = None,
    ) -> str:
        # This method is an exception to the normal exception-handling to fail over to the next RPC provider.
        #
        # Normal RPC exceptions just move on to the next RPC provider and try again. That won't work with the
        # BlockhashNotFoundException, since a stale blockhash will be stale for all providers and it probably
        # indicates a problem with the current node returning the stale blockhash anyway.
        #
        # What we want to do in this situation is: retry the same transaction (which we know for certain failed)
        # but retry it with the next provider in the list, with a fresh recent_blockhash. (Setting the transaction's
        # recent_blockhash to None makes the client fetch a fresh one.)
        last_exception: BlockhashNotFoundException
        for provider in self.rpc_caller.all_providers:
            try:
                proper_commitment: Commitment = opts.preflight_commitment
                proper_skip_preflight = opts.skip_preflight
                proper_tpu_retransmissions = opts.max_retries
                if proper_commitment == UnspecifiedCommitment:
                    proper_commitment = self.commitment
                    proper_skip_preflight = self.skip_preflight
                    proper_tpu_retransmissions = (
                        self.tpu_retransmissions
                        if self.tpu_retransmissions >= 0
                        else None
                    )

                proper_opts = TxOpts(
                    preflight_commitment=proper_commitment,
                    skip_confirmation=opts.skip_confirmation,
                    skip_preflight=proper_skip_preflight,
                    max_retries=proper_tpu_retransmissions,
                )

                response = self.compatible_client.send_transaction(
                    transaction,
                    *signers,
                    opts=proper_opts,
                    recent_blockhash=recent_blockhash,
                )
                signature: str = str(response["result"])
                self._logger.debug(f"Transaction signature: {signature}")

                if signature != _STUB_TRANSACTION_SIGNATURE:
                    self.transaction_monitor.monitor(signature)
                else:
                    self._logger.error("Could not get status for stub signature")

                return signature
            except BlockhashNotFoundException as blockhash_not_found_exception:
                self._logger.debug(
                    f"Trying next provider after intercepting blockhash exception on provider {provider}: {blockhash_not_found_exception}"
                )
                last_exception = blockhash_not_found_exception
                transaction.recent_blockhash = None
                self.rpc_caller.shift_to_next_provider()

        raise last_exception

    def wait_for_confirmation(
        self, transaction_ids: typing.Sequence[str], max_wait_in_seconds: int = 60
    ) -> typing.Sequence[str]:
        self._logger.info(
            f"Waiting up to {max_wait_in_seconds} seconds for {transaction_ids}."
        )
        all_confirmed: typing.List[str] = []
        start_time: datetime = local_now()
        cutoff: datetime = start_time + timedelta(seconds=max_wait_in_seconds)
        for transaction_id in transaction_ids:
            while local_now() < cutoff:
                time.sleep(1)
                confirmed = self.get_confirmed_transaction(transaction_id)
                if confirmed is not None:
                    self._logger.info(
                        f"Confirmed {transaction_id} after {local_now() - start_time} seconds."
                    )
                    all_confirmed += [transaction_id]
                    break

        if len(all_confirmed) != len(transaction_ids):
            self._logger.info(
                f"Timed out after {max_wait_in_seconds} seconds waiting on transaction {transaction_id}."
            )
        return all_confirmed

    def __resolve_defaults(
        self,
        commitment: typing.Optional[Commitment],
        encoding: typing.Optional[str] = None,
    ) -> typing.Tuple[Commitment, str]:
        if commitment is None or commitment == UnspecifiedCommitment:
            commitment = self.commitment

        if encoding is None or encoding == UnspecifiedEncoding:
            encoding = self.encoding

        return commitment, encoding

    def __str__(self) -> str:
        return f"« BetterClient [{self.cluster_name}]: {self.cluster_rpc_urls}, {self.cluster_ws_urls} »"

    def __repr__(self) -> str:
        return f"{self}"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)


import typing

from .placedorder import PlacedOrder


# # 🥭 PerpOpenOrders class
#
class PerpOpenOrders:
    def __init__(self, placed_orders: typing.Sequence[PlacedOrder]) -> None:
        self.placed_orders: typing.Sequence[PlacedOrder] = placed_orders

    @property
    def empty(self) -> bool:
        return len(self.placed_orders) == 0

    def __str__(self) -> str:
        placed_orders = "\n        ".join(map(str, self.placed_orders)) or "None"

        return f"""« PerpOpenOrders
    Orders:
        {placed_orders}
»"""

    def __repr__(self) -> str:
        return f"{self}"



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import construct
import typing

from solana.publickey import PublicKey

from ...layouts.layouts import DecimalAdapter, PublicKeyAdapter


# # 🥭 Pyth
#
# This file contains code specific to the [Pyth Network](https://pyth.network/).
#

# # 🥭 Constants
#
# Pyth defines some constants.
#

MAGIC = 0xA1B2C3D4
VERSION_2 = 2
VERSION = VERSION_2
MAP_TABLE_SIZE = 640
PROD_ACCT_SIZE = 512
PROD_HDR_SIZE = 48
PROD_ATTR_SIZE = PROD_ACCT_SIZE - PROD_HDR_SIZE

PYTH_DEVNET_MAPPING_ROOT = PublicKey("BmA9Z6FjioHJPpjT39QazZyhDRUdZy2ezwx4GiDdE2u2")
PYTH_MAINNET_MAPPING_ROOT = PublicKey("AHtgzX45WTKfkPG53L6WYhGEXwQkN1BVknET3sVsLL8J")


# # 🥭 ACCOUNT_TYPE enum
#
# This layout implements the AccountType enum. Here's the [Rust implementation](https://github.com/pyth-network/pyth-client-rs/blob/main/src/lib.rs).
#
# // each account has its own type
# #[repr(C)]
# pub enum AccountType
# {
#   Unknown,
#   Mapping,
#   Product,
#   Price
# }

ACCOUNT_TYPE = construct.Enum(
    construct.Int32ul, Unknown=0, Mapping=1, Product=2, Price=3
)


# # 🥭 PythStringDictionaryAdapter
#
# Loads string key-value pairs into a dictionary. The strings are 'Pascal' strings, which
# specify the string length as an int followed by that number of UTF-8 characters.
#
if typing.TYPE_CHECKING:

    class StringDictionaryAdapter(
        construct.Adapter[typing.Any, str, typing.Any, typing.Any]
    ):
        def __init__(self) -> None:
            pass

else:

    class StringDictionaryAdapter(construct.Adapter):
        def __init__(self) -> None:
            super().__init__(
                construct.RepeatUntil(
                    lambda contents, lst, ctx: contents == "",
                    construct.PascalString(construct.VarInt, "utf8"),
                )
            )

        def _decode(
            self, obj: typing.Sequence[str], context: typing.Any, path: typing.Any
        ) -> typing.Dict[str, str]:
            result: typing.Dict[str, str] = {}
            for counter in range(int(len(obj) / 2)):
                index: int = counter * 2
                key: str = obj[index]
                value: str = obj[index + 1]
                result[key] = value
            return result

        def _encode(
            self, obj: typing.Any, context: typing.Any, path: typing.Any
        ) -> str:
            # Can only encode string values.
            return str(obj)


# # 🥭 MAPPING layout
#
# This layout implements the Mapping structure. Here's the [Rust implementation](https://github.com/pyth-network/pyth-client-rs/blob/main/src/lib.rs).
#
# // Mapping account structure
# #[repr(C)]
# pub struct Mapping
# {
#   pub magic      : u32,        // pyth magic number
#   pub ver        : u32,        // program version
#   pub atype      : u32,        // account type
#   pub size       : u32,        // account used size
#   pub num        : u32,        // number of product accounts
#   pub unused     : u32,
#   pub next       : AccKey,     // next mapping account (if any)
#   pub products   : [AccKey;MAP_TABLE_SIZE]
# }
MAPPING = construct.Struct(
    "magic" / DecimalAdapter(4),
    "ver" / DecimalAdapter(4),
    "atype" / ACCOUNT_TYPE,
    "size" / DecimalAdapter(4),
    "num" / DecimalAdapter(4),
    "unused" / DecimalAdapter(4),
    "next" / PublicKeyAdapter(),
    "products" / construct.Array(MAP_TABLE_SIZE, PublicKeyAdapter()),
)


# # 🥭 PRODUCT layout
#
# This layout implements the Product structure. Here's the [Rust implementation](https://github.com/pyth-network/pyth-client-rs/blob/main/src/lib.rs).
#
#
# // Product account structure
# #[repr(C)]
# pub struct Product
# {
#   pub magic      : u32,        // pyth magic number
#   pub ver        : u32,        // program version
#   pub atype      : u32,        // account type
#   pub size       : u32,        // price account size
#   pub px_acc     : AccKey,     // first price account in list
#   pub attr       : [u8;PROD_ATTR_SIZE] // key/value pairs of reference attr.
# }
PRODUCT = construct.Struct(
    "magic" / DecimalAdapter(4),
    "ver" / DecimalAdapter(4),
    "atype" / ACCOUNT_TYPE,
    "size" / DecimalAdapter(4),
    "px_acc" / PublicKeyAdapter(),
    "attr" / StringDictionaryAdapter(),
)


# # 🥭 PRICE_INFO layout
#
# This layout implements the PriceInfo structure. Here's the [Rust implementation](https://github.com/pyth-network/pyth-client-rs/blob/main/src/lib.rs).
#
# // contributing or aggregate price component
# #[repr(C)]
# pub struct PriceInfo
# {
#   pub price      : i64,        // product price
#   pub conf       : u64,        // confidence interval of product price
#   pub status     : PriceStatus,// status of price (Trading is valid)
#   pub corp_act   : CorpAction, // notification of any corporate action
#   pub pub_slot   : u64
# }
PRICE_INFO = construct.Struct(
    "price" / DecimalAdapter(),
    "conf" / DecimalAdapter(),
    "status"
    / construct.Enum(construct.Int32ul, Unknown=0, Trading=1, Halted=2, Auction=3),
    "corp_act" / construct.Enum(construct.Int32ul, NoCorpAct=0),
    "pub_slot" / DecimalAdapter(),
)


# # 🥭 PRICE_COMP layout
#
# This layout implements the PriceComp structure. Here's the [Rust implementation](https://github.com/pyth-network/pyth-client-rs/blob/main/src/lib.rs).
#
# // latest component price and price used in aggregate snapshot
# #[repr(C)]
# pub struct PriceComp
# {
#   publisher  : AccKey,         // key of contributing quoter
#   agg        : PriceInfo,      // contributing price to last aggregate
#   latest     : PriceInfo       // latest contributing price (not in agg.)
# }
PRICE_COMP = construct.Struct(
    "publisher" / PublicKeyAdapter(), "agg" / PRICE_INFO, "latest" / PRICE_INFO
)


# # 🥭 PRICE layout
#
# This layout implements the Price structure. Here's the [Rust implementation](https://github.com/pyth-network/pyth-client-rs/blob/main/src/lib.rs).
#
# // Price account structure V2
# #[repr(C)]
# pub struct Price
# {
#   pub magic      : u32,        // pyth magic number
#   pub ver        : u32,        // program version
#   pub atype      : u32,        // account type
#   pub size       : u32,        // price account size
#   pub ptype      : PriceType,  // price or calculation type
#   pub expo       : i32,        // price exponent
#   pub num        : u32,        // number of component prices
#   pub unused     : u32,
#   pub curr_slot  : u64,        // currently accumulating price slot
#   pub valid_slot : u64,        // valid slot-time of agg. price
#   pub drv        : [i64;8],    // calculated values derived from agg. price
#   pub prod       : AccKey,     // product account key
#   pub next       : AccKey,     // next Price account in linked list
#   pub agg_pub    : AccKey,     // quoter who computed last aggregate price
#   pub agg        : PriceInfo,  // aggregate price info
#   pub comp       : [PriceComp;32] // price components one per quoter
# }
PRICE = construct.Struct(
    "magic" / DecimalAdapter(4),
    "ver" / DecimalAdapter(4),
    "atype" / ACCOUNT_TYPE,
    "size" / DecimalAdapter(4),
    "ptype" / construct.Enum(construct.Int32ul, Unknown=0, Price=1),
    "expo" / construct.Int32sl,
    "num" / DecimalAdapter(4),
    "unused" / DecimalAdapter(4),
    "curr_slot" / DecimalAdapter(),
    "valid_slot" / DecimalAdapter(),
    "drv" / construct.Array(8, construct.Int64sl),
    "prod" / PublicKeyAdapter(),
    "next" / PublicKeyAdapter(),
    "agg_pub" / PublicKeyAdapter(),
    "agg" / PRICE_INFO,
    "comp" / construct.Array(32, PRICE_COMP),
)



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

from .context import Context
from .contextbuilder import ContextBuilder
from .oracle import OracleProvider
from .oracles.ftx import ftx
from .oracles.market import market
from .oracles.pythnetwork import pythnetwork
from .oracles.stub import stub


# # 🥭 Oracle Factory
#
# This file allows you to create a concreate OracleProvider for a specified provider name.
#
def create_oracle_provider(context: Context, provider_name: str) -> OracleProvider:
    proper_provider_name: str = provider_name.upper()
    if proper_provider_name == "FTX":
        return ftx.FtxOracleProvider()
    elif proper_provider_name == "MARKET":
        return market.MarketOracleProvider()
    elif proper_provider_name == "PYTH":
        return pythnetwork.PythOracleProvider(context)
    elif proper_provider_name == "PYTH-MAINNET":
        mainnet_beta_pyth_context: Context = ContextBuilder.force_to_cluster(
            context, "mainnet"
        )
        return pythnetwork.PythOracleProvider(mainnet_beta_pyth_context)
    elif proper_provider_name == "PYTH-DEVNET":
        devnet_pyth_context: Context = ContextBuilder.force_to_cluster(
            context, "devnet"
        )
        return pythnetwork.PythOracleProvider(devnet_pyth_context)
    elif proper_provider_name == "STUB":
        return stub.StubOracleProvider()
    raise Exception(f"Unknown oracle provider '{proper_provider_name}'.")



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

from decimal import Decimal

from .account import Account
from .context import Context
from .group import Group
from .instrumentvalue import InstrumentValue
from .loadedmarket import LoadedMarket
from .marketoperations import (
    MarketInstructionBuilder,
    MarketOperations,
    NullMarketInstructionBuilder,
    NullMarketOperations,
)
from .perpmarket import (
    PerpMarketStub,
    PerpMarket,
    PerpMarketInstructionBuilder,
    PerpMarketOperations,
)
from .serummarket import (
    SerumMarketStub,
    SerumMarket,
    SerumMarketInstructionBuilder,
    SerumMarketOperations,
)
from .spotmarket import (
    SpotMarketStub,
    SpotMarket,
    SpotMarketInstructionBuilder,
    SpotMarketOperations,
)
from .tokens import Instrument, Token
from .wallet import Wallet


# # 🥭 instrument
#
# This function takes a `Context` and an instrument symbol string (like "ETH") and returns a
# properly loaded `Instrument` object. It throws if anything goes wrong rather than return None.
#
# An Instrument can represent a Perp or an SPL token.
#
def instrument(context: Context, symbol: str) -> Instrument:
    return context.instrument(symbol)


# # 🥭 token
#
# This function takes a `Context` and a token symbol string (like "ETH") and returns a
# properly loaded `Token` object. It throws if anything goes wrong rather than return None.
#
# An Token can only represent an SPL token. It cannot represent a Perp.
#
def token(context: Context, symbol: str) -> Token:
    return context.token(symbol)


# # 🥭 instrument_value
#
# This function takes a `Context` and an instrument symbol string (like "ETH") and returns a
# properly loaded `Instrument` object. It throws if anything goes wrong rather than return None.
#
# An Instrument can represent a Perp or an SPL token.
#
def instrument_value(context: Context, symbol: str, value: Decimal) -> InstrumentValue:
    ins = context.instrument(symbol)
    return InstrumentValue(ins, value)


# # 🥭 market
#
# This function takes a `Context` and a market symbol string (like "ETH/USDC") and returns a
# properly loaded market. It throws if anything goes wrong rather than return None.
#
def market(context: Context, symbol: str) -> LoadedMarket:
    market = context.market_lookup.find_by_symbol(symbol)
    if market is None:
        raise Exception(f"Could not find market {symbol}")

    if isinstance(market, LoadedMarket):
        return market
    elif isinstance(market, SerumMarketStub):
        return market.load(context)
    elif isinstance(market, SpotMarketStub):
        group: Group = Group.load(context, market.group_address)
        return market.load(context, group)
    elif isinstance(market, PerpMarketStub):
        group = Group.load(context, market.group_address)
        return market.load(context, group)

    raise Exception(f"Market {market} could not be loaded.")


# # 🥭 instruction_builder
#
# This function deals with the creation of a `MarketInstructionBuilder` object for a given
# market.
#
def instruction_builder(
    context: Context,
    wallet: Wallet,
    account: Account,
    symbol: str,
    dry_run: bool = False,
) -> MarketInstructionBuilder:
    loaded_market: LoadedMarket = market(context, symbol)
    if dry_run:
        return NullMarketInstructionBuilder(loaded_market.fully_qualified_symbol)

    if SerumMarket.isa(loaded_market):
        return SerumMarketInstructionBuilder.load(
            context, wallet, SerumMarket.ensure(loaded_market)
        )
    elif SpotMarket.isa(loaded_market):
        spot_market = SpotMarket.ensure(loaded_market)
        return SpotMarketInstructionBuilder.load(
            context,
            wallet,
            spot_market,
            spot_market.group,
            account,
        )
    elif PerpMarket.isa(loaded_market):
        perp_market = PerpMarket.ensure(loaded_market)
        return PerpMarketInstructionBuilder.load(
            context,
            wallet,
            perp_market,
            perp_market.group,
            account,
        )

    raise Exception(f"Could not find instructions builder for market {symbol}")


# # 🥭 operations
#
# This function deals with the creation of a `MarketOperations` object for a given market.
#
def operations(
    context: Context,
    wallet: Wallet,
    account: Account,
    symbol: str,
    dry_run: bool = False,
) -> MarketOperations:
    loaded_market: LoadedMarket = market(context, symbol)
    if dry_run:
        return NullMarketOperations(loaded_market)

    if SerumMarket.isa(loaded_market):
        serum_market_instruction_builder: SerumMarketInstructionBuilder = (
            SerumMarketInstructionBuilder.load(
                context, wallet, SerumMarket.ensure(loaded_market)
            )
        )
        return SerumMarketOperations(context, wallet, serum_market_instruction_builder)
    elif SpotMarket.isa(loaded_market):
        spot_market = SpotMarket.ensure(loaded_market)
        spot_market_instruction_builder: SpotMarketInstructionBuilder = (
            SpotMarketInstructionBuilder.load(
                context, wallet, spot_market, spot_market.group, account
            )
        )
        return SpotMarketOperations(
            context, wallet, account, spot_market_instruction_builder
        )
    elif PerpMarket.isa(loaded_market):
        perp_market = PerpMarket.ensure(loaded_market)
        perp_market_instruction_builder: PerpMarketInstructionBuilder = (
            PerpMarketInstructionBuilder.load(
                context,
                wallet,
                perp_market,
                perp_market.underlying_perp_market.group,
                account,
            )
        )
        return PerpMarketOperations(
            context, wallet, account, perp_market_instruction_builder
        )

    raise Exception(f"Could not find operations for market {symbol}")



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import logging
import typing

from decimal import Decimal
from pyserum.market.market import Market as PySerumMarket
from solana.publickey import PublicKey

from .account import Account
from .accountinfo import AccountInfo
from .cache import Cache
from .combinableinstructions import CombinableInstructions
from .context import Context
from .group import GroupSlot, Group
from .healthcheck import HealthCheck
from .instructions import build_serum_create_openorders_instructions
from .instrumentvalue import InstrumentValue
from .inventory import Inventory
from .loadedmarket import LoadedMarket
from .markets import InventorySource
from .modelstate import EventQueue
from .observables import Disposable, LatestItemObserverSubscriber
from .openorders import OpenOrders
from .oracle import Price
from .oracle import OracleProvider
from .oraclefactory import create_oracle_provider
from .orders import OrderBook
from .perpeventqueue import PerpEventQueue
from .perpmarket import PerpMarket
from .placedorder import PlacedOrdersContainer
from .serumeventqueue import SerumEventQueue
from .serummarket import SerumMarket
from .spotmarket import SpotMarket, SpotMarketInstructionBuilder, SpotMarketOperations
from .tokenaccount import TokenAccount
from .tokens import Instrument, Token
from .wallet import Wallet
from .watcher import Watcher, LamdaUpdateWatcher
from .websocketsubscription import (
    WebSocketAccountSubscription,
    WebSocketSubscription,
    WebSocketSubscriptionManager,
)


def build_group_watcher(
    context: Context,
    manager: WebSocketSubscriptionManager,
    health_check: HealthCheck,
    group: Group,
) -> Watcher[Group]:
    group_subscription = WebSocketAccountSubscription[Group](
        context,
        group.address,
        lambda account_info: Group.parse(
            account_info, group.name, context.instrument_lookup, context.market_lookup
        ),
    )
    manager.add(group_subscription)
    latest_group_observer = LatestItemObserverSubscriber[Group](group)
    group_subscription.publisher.subscribe(latest_group_observer)
    health_check.add("group_subscription", group_subscription.publisher)
    return latest_group_observer


def build_account_watcher(
    context: Context,
    manager: WebSocketSubscriptionManager,
    health_check: HealthCheck,
    account: Account,
    group_observer: Watcher[Group],
    cache_observer: Watcher[Cache],
) -> typing.Tuple[WebSocketSubscription[Account], Watcher[Account]]:
    account_subscription = WebSocketAccountSubscription[Account](
        context,
        account.address,
        lambda account_info: Account.parse(
            account_info, group_observer.latest, cache_observer.latest
        ),
    )
    manager.add(account_subscription)
    latest_account_observer = LatestItemObserverSubscriber[Account](account)
    account_subscription.publisher.subscribe(latest_account_observer)
    health_check.add("account_subscription", account_subscription.publisher)
    return account_subscription, latest_account_observer


def build_cache_watcher(
    context: Context,
    manager: WebSocketSubscriptionManager,
    health_check: HealthCheck,
    cache: Cache,
    group: Group,
) -> Watcher[Cache]:
    cache_subscription = WebSocketAccountSubscription[Cache](
        context, group.cache, lambda account_info: Cache.parse(account_info)
    )
    manager.add(cache_subscription)
    latest_cache_observer = LatestItemObserverSubscriber[Cache](cache)
    cache_subscription.publisher.subscribe(latest_cache_observer)
    health_check.add("cache_subscription", cache_subscription.publisher)
    return latest_cache_observer


def build_spot_open_orders_watcher(
    context: Context,
    manager: WebSocketSubscriptionManager,
    health_check: HealthCheck,
    wallet: Wallet,
    account: Account,
    group: Group,
    spot_market: SpotMarket,
) -> Watcher[OpenOrders]:
    market_index = group.slot_by_spot_market_address(spot_market.address).index
    open_orders_address = account.spot_open_orders_by_index[market_index]
    if open_orders_address is None:
        spot_market_instruction_builder: SpotMarketInstructionBuilder = (
            SpotMarketInstructionBuilder.load(
                context, wallet, spot_market, spot_market.group, account
            )
        )
        market_operations: SpotMarketOperations = SpotMarketOperations(
            context, wallet, account, spot_market_instruction_builder
        )
        open_orders_address = market_operations.create_openorders()
        logging.info(
            f"Created {spot_market.fully_qualified_symbol} OpenOrders at: {open_orders_address}"
        )

    spot_open_orders_subscription = WebSocketAccountSubscription[OpenOrders](
        context,
        open_orders_address,
        lambda account_info: OpenOrders.parse(
            account_info, spot_market.base, spot_market.quote
        ),
    )
    manager.add(spot_open_orders_subscription)
    initial_spot_open_orders = OpenOrders.load(
        context,
        open_orders_address,
        spot_market.base,
        spot_market.quote,
    )
    latest_open_orders_observer = LatestItemObserverSubscriber[OpenOrders](
        initial_spot_open_orders
    )
    spot_open_orders_subscription.publisher.subscribe(latest_open_orders_observer)
    health_check.add(
        "open_orders_subscription", spot_open_orders_subscription.publisher
    )
    return latest_open_orders_observer


def build_serum_open_orders_watcher(
    context: Context,
    manager: WebSocketSubscriptionManager,
    health_check: HealthCheck,
    serum_market: SerumMarket,
    wallet: Wallet,
) -> Watcher[PlacedOrdersContainer]:
    all_open_orders = OpenOrders.load_for_market_and_owner(
        context,
        serum_market.address,
        wallet.address,
        context.serum_program_address,
        serum_market.base,
        serum_market.quote,
    )
    if len(all_open_orders) > 0:
        initial_serum_open_orders: OpenOrders = all_open_orders[0]
        open_orders_address = initial_serum_open_orders.address
    else:
        raw_market = PySerumMarket.load(
            context.client.compatible_client, serum_market.address
        )
        create_open_orders = build_serum_create_openorders_instructions(
            context, wallet, raw_market
        )

        open_orders_address = create_open_orders.signers[0].public_key

        logging.info(
            f"Creating OpenOrders account for market {serum_market.fully_qualified_symbol} at {open_orders_address}."
        )
        signers: CombinableInstructions = CombinableInstructions.from_wallet(wallet)
        transaction_ids = (signers + create_open_orders).execute(context)
        context.client.wait_for_confirmation(transaction_ids)
        initial_serum_open_orders = OpenOrders.load(
            context,
            open_orders_address,
            serum_market.base,
            serum_market.quote,
        )

    serum_open_orders_subscription = WebSocketAccountSubscription[OpenOrders](
        context,
        open_orders_address,
        lambda account_info: OpenOrders.parse(
            account_info, serum_market.base, serum_market.quote
        ),
    )

    manager.add(serum_open_orders_subscription)

    latest_serum_open_orders_observer = LatestItemObserverSubscriber[
        PlacedOrdersContainer
    ](initial_serum_open_orders)
    serum_open_orders_subscription.publisher.subscribe(
        latest_serum_open_orders_observer
    )
    health_check.add(
        "open_orders_subscription", serum_open_orders_subscription.publisher
    )
    return latest_serum_open_orders_observer


def build_perp_open_orders_watcher(
    context: Context,
    manager: WebSocketSubscriptionManager,
    health_check: HealthCheck,
    perp_market: PerpMarket,
    account: Account,
    group: Group,
    account_subscription: WebSocketSubscription[Account],
) -> Watcher[PlacedOrdersContainer]:
    slot: GroupSlot = group.slot_by_perp_market_address(perp_market.address)
    index: int = slot.index
    initial_perp_account = account.perp_accounts_by_index[slot.index]
    if initial_perp_account is None:
        raise Exception(
            f"Could not find perp account at index {slot.index} of account {account.address}."
        )
    initial_open_orders = initial_perp_account.open_orders
    latest_open_orders_observer = LatestItemObserverSubscriber[PlacedOrdersContainer](
        initial_open_orders
    )
    account_subscription.publisher.subscribe(
        on_next=lambda updated_account: latest_open_orders_observer.on_next(
            updated_account.perp_accounts_by_index[index].open_orders
        )
    )  # type: ignore[call-arg]
    health_check.add("open_orders_subscription", account_subscription.publisher)
    return latest_open_orders_observer


def build_price_watcher(
    context: Context,
    manager: WebSocketSubscriptionManager,
    health_check: HealthCheck,
    disposer: Disposable,
    provider_name: str,
    market: LoadedMarket,
) -> LatestItemObserverSubscriber[Price]:
    oracle_provider: OracleProvider = create_oracle_provider(context, provider_name)
    oracle = oracle_provider.oracle_for_market(context, market)
    if oracle is None:
        raise Exception(
            f"Could not find oracle for market {market.fully_qualified_symbol} from provider {provider_name}."
        )

    initial_price = oracle.fetch_price(context)
    price_feed = oracle.to_streaming_observable(context)
    latest_price_observer = LatestItemObserverSubscriber(initial_price)
    price_disposable = price_feed.subscribe(latest_price_observer)
    disposer.add_disposable(price_disposable)
    health_check.add("price_subscription", price_feed)
    return latest_price_observer


def build_serum_inventory_watcher(
    context: Context,
    manager: WebSocketSubscriptionManager,
    health_check: HealthCheck,
    disposer: Disposable,
    wallet: Wallet,
    market: SerumMarket,
    price_watcher: Watcher[Price],
) -> Watcher[Inventory]:
    base_account = TokenAccount.fetch_largest_for_owner_and_token(
        context, wallet.address, market.base
    )
    if base_account is None:
        raise Exception(
            f"Could not find token account owned by {wallet.address} for base token {market.base}."
        )
    base_token_subscription = WebSocketAccountSubscription[TokenAccount](
        context,
        base_account.address,
        lambda account_info: TokenAccount.parse(account_info, market.base),
    )
    manager.add(base_token_subscription)
    latest_base_token_account_observer = LatestItemObserverSubscriber[TokenAccount](
        base_account
    )
    base_subscription_disposable = base_token_subscription.publisher.subscribe(
        latest_base_token_account_observer
    )
    disposer.add_disposable(base_subscription_disposable)

    quote_account = TokenAccount.fetch_largest_for_owner_and_token(
        context, wallet.address, market.quote
    )
    if quote_account is None:
        raise Exception(
            f"Could not find token account owned by {wallet.address} for quote token {market.quote}."
        )
    quote_token_subscription = WebSocketAccountSubscription[TokenAccount](
        context,
        quote_account.address,
        lambda account_info: TokenAccount.parse(account_info, market.quote),
    )
    manager.add(quote_token_subscription)
    latest_quote_token_account_observer = LatestItemObserverSubscriber[TokenAccount](
        quote_account
    )
    quote_subscription_disposable = quote_token_subscription.publisher.subscribe(
        latest_quote_token_account_observer
    )
    disposer.add_disposable(quote_subscription_disposable)

    # Serum markets don't accrue MNGO liquidity incentives
    mngo: typing.Optional[Instrument] = context.instrument_lookup.find_by_symbol("MNGO")
    if mngo is None:
        raise Exception("Could not find details of MNGO token.")
    mngo_accrued: InstrumentValue = InstrumentValue(Token.ensure(mngo), Decimal(0))

    def serum_inventory_accessor() -> Inventory:
        available: Decimal = (
            latest_base_token_account_observer.latest.value.value
            * price_watcher.latest.mid_price
        ) + latest_quote_token_account_observer.latest.value.value
        available_collateral: InstrumentValue = InstrumentValue(
            latest_quote_token_account_observer.latest.value.token, available
        )
        return Inventory(
            InventorySource.SPL_TOKENS,
            mngo_accrued,
            available_collateral,
            latest_base_token_account_observer.latest.value,
            latest_quote_token_account_observer.latest.value,
        )

    return LamdaUpdateWatcher(serum_inventory_accessor)


def build_orderbook_watcher(
    context: Context,
    manager: WebSocketSubscriptionManager,
    health_check: HealthCheck,
    market: LoadedMarket,
) -> Watcher[OrderBook]:
    orderbook_addresses: typing.List[PublicKey] = [
        market.bids_address,
        market.asks_address,
    ]
    orderbook_infos = AccountInfo.load_multiple(context, orderbook_addresses)
    if (
        len(orderbook_infos) != 2
        or orderbook_infos[0] is None
        or orderbook_infos[1] is None
    ):
        raise Exception(
            f"Could not find {market.fully_qualified_symbol} order book at addresses {orderbook_addresses}."
        )

    initial_orderbook: OrderBook = market.parse_account_infos_to_orderbook(
        orderbook_infos[0], orderbook_infos[1]
    )
    updatable_orderbook: OrderBook = market.parse_account_infos_to_orderbook(
        orderbook_infos[0], orderbook_infos[1]
    )

    def _update_bids(account_info: AccountInfo) -> OrderBook:
        new_bids = market.parse_account_info_to_orders(account_info)
        updatable_orderbook.bids = new_bids
        return updatable_orderbook

    def _update_asks(account_info: AccountInfo) -> OrderBook:
        new_asks = market.parse_account_info_to_orders(account_info)
        updatable_orderbook.asks = new_asks
        return updatable_orderbook

    bids_subscription = WebSocketAccountSubscription[OrderBook](
        context, orderbook_addresses[0], _update_bids
    )
    manager.add(bids_subscription)
    asks_subscription = WebSocketAccountSubscription[OrderBook](
        context, orderbook_addresses[1], _update_asks
    )
    manager.add(asks_subscription)

    orderbook_observer = LatestItemObserverSubscriber[OrderBook](initial_orderbook)

    bids_subscription.publisher.subscribe(orderbook_observer)
    asks_subscription.publisher.subscribe(orderbook_observer)
    health_check.add("orderbook_bids_subscription", bids_subscription.publisher)
    health_check.add("orderbook_asks_subscription", asks_subscription.publisher)
    return orderbook_observer


def build_serum_event_queue_watcher(
    context: Context,
    manager: WebSocketSubscriptionManager,
    health_check: HealthCheck,
    serum_market: SerumMarket,
) -> Watcher[EventQueue]:
    initial: EventQueue = SerumEventQueue.load(
        context, serum_market.event_queue_address, serum_market.base, serum_market.quote
    )
    subscription = WebSocketAccountSubscription[EventQueue](
        context,
        serum_market.event_queue_address,
        lambda account_info: SerumEventQueue.parse(
            account_info, serum_market.base, serum_market.quote
        ),
    )
    manager.add(subscription)
    latest_observer = LatestItemObserverSubscriber[EventQueue](initial)
    subscription.publisher.subscribe(latest_observer)
    health_check.add("event_queue_subscription", subscription.publisher)
    return latest_observer


def build_spot_event_queue_watcher(
    context: Context,
    manager: WebSocketSubscriptionManager,
    health_check: HealthCheck,
    spot_market: SpotMarket,
) -> Watcher[EventQueue]:
    initial: EventQueue = SerumEventQueue.load(
        context, spot_market.event_queue_address, spot_market.base, spot_market.quote
    )
    subscription = WebSocketAccountSubscription[EventQueue](
        context,
        spot_market.event_queue_address,
        lambda account_info: SerumEventQueue.parse(
            account_info, spot_market.base, spot_market.quote
        ),
    )
    manager.add(subscription)
    latest_observer = LatestItemObserverSubscriber[EventQueue](initial)
    subscription.publisher.subscribe(latest_observer)
    health_check.add("event_queue_subscription", subscription.publisher)
    return latest_observer


def build_perp_event_queue_watcher(
    context: Context,
    manager: WebSocketSubscriptionManager,
    health_check: HealthCheck,
    perp_market: PerpMarket,
) -> Watcher[EventQueue]:
    initial: EventQueue = PerpEventQueue.load(
        context, perp_market.event_queue_address, perp_market.lot_size_converter
    )
    subscription = WebSocketAccountSubscription[EventQueue](
        context,
        perp_market.event_queue_address,
        lambda account_info: PerpEventQueue.parse(
            account_info, perp_market.lot_size_converter
        ),
    )
    manager.add(subscription)
    latest_observer = LatestItemObserverSubscriber[EventQueue](initial)
    subscription.publisher.subscribe(latest_observer)
    health_check.add("event_queue_subscription", subscription.publisher)
    return latest_observer



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import abc
import logging
import typing
import websocket

from dataclasses import dataclass
from rx.subject.behaviorsubject import BehaviorSubject
from rx.core.typing import Disposable as RxDisposable
from solana.publickey import PublicKey
from solana.rpc.types import RPCResponse

from .accountinfo import AccountInfo
from .context import Context
from .datetimes import local_now
from .observables import EventSource
from .reconnectingwebsocket import ReconnectingWebsocket


# # 🥭 WebSocketSubscription class
#
# The `WebSocketSubscription` maintains a mapping for an account subscription in a Solana websocket to
# an actual instantiated object.
#


TSubscriptionInstance = typing.TypeVar("TSubscriptionInstance")


class WebSocketSubscription(
    RxDisposable, typing.Generic[TSubscriptionInstance], metaclass=abc.ABCMeta
):
    def __init__(
        self,
        context: Context,
    ) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.context: Context = context
        self.id: int = context.generate_client_id()
        self.subscription_id: int = 0
        self.publisher: EventSource[TSubscriptionInstance] = EventSource[
            TSubscriptionInstance
        ]()

    @abc.abstractmethod
    def build_request(self) -> str:
        raise NotImplementedError(
            "WebSocketSubscription.build_request() is not implemented on the base type."
        )

    def _on_item(self, response: typing.Dict[str, typing.Any]) -> None:
        if "method" not in response:
            id: int = int(response["id"])
            if id == self.id:
                subscription_id: int = int(response["result"])
                self._logger.info(f"Subscription created with id {subscription_id}.")
        else:
            subscription_id = response["params"]["subscription"]
            built = self.build_subscribed_instance(response["params"])
            self.publisher.publish(built)

    @abc.abstractmethod
    def build_subscribed_instance(self, response: RPCResponse) -> TSubscriptionInstance:
        raise NotImplementedError(
            "WebSocketSubscription.build_subscribed_instance() is not implemented on the base type."
        )

    def dispose(self) -> None:
        self.publisher.on_completed()


class AddressWebSocketSubscription(WebSocketSubscription[TSubscriptionInstance]):
    def __init__(
        self,
        context: Context,
        address: PublicKey,
        constructor: typing.Callable[[AccountInfo], TSubscriptionInstance],
    ) -> None:
        super().__init__(context)
        self.address: PublicKey = address
        self.from_account_info: typing.Callable[
            [AccountInfo], TSubscriptionInstance
        ] = constructor

    def build_subscribed_instance(self, response: RPCResponse) -> TSubscriptionInstance:
        account_info: AccountInfo = AccountInfo.from_response(response, self.address)
        built: TSubscriptionInstance = self.from_account_info(account_info)
        return built


class WebSocketProgramSubscription(AddressWebSocketSubscription[TSubscriptionInstance]):
    def __init__(
        self,
        context: Context,
        address: PublicKey,
        constructor: typing.Callable[[AccountInfo], TSubscriptionInstance],
    ) -> None:
        super().__init__(context, address, constructor)

    def build_request(self) -> str:
        return f"""{{
    "jsonrpc": "2.0",
    "id": {self.id},
    "method": "programSubscribe",
    "params": [
        "{self.address}",
        {{
            "encoding": "base64",
            "commitment": "{self.context.client.commitment}"
        }}
    ]
}}"""


class WebSocketAccountSubscription(AddressWebSocketSubscription[TSubscriptionInstance]):
    def __init__(
        self,
        context: Context,
        address: PublicKey,
        constructor: typing.Callable[[AccountInfo], TSubscriptionInstance],
    ) -> None:
        super().__init__(context, address, constructor)

    def build_request(self) -> str:
        return f"""{{
    "jsonrpc": "2.0",
    "id": {self.id},
    "method": "accountSubscribe",
    "params": [
        "{self.address}",
        {{
            "encoding": "base64",
            "commitment": "{self.context.client.commitment}"
        }}
    ]
}}"""


class WebSocketSignatureSubscription(WebSocketSubscription[RPCResponse]):
    def __init__(
        self,
        context: Context,
        signature: str,
    ) -> None:
        super().__init__(context)
        self.signature: str = signature

    def build_request(self) -> str:
        return f"""{{
    "jsonrpc": "2.0",
    "id": {self.id},
    "method": "signatureSubscribe",
    "params": [
        "{self.signature}",
        {{
            "commitment": "finalized"
        }}
    ]
}}"""

    def build_subscribed_instance(self, response: RPCResponse) -> RPCResponse:
        return response


class LogEvent:
    def __init__(
        self, signatures: typing.Sequence[str], logs: typing.Sequence[str]
    ) -> None:
        self.signatures: typing.Sequence[str] = signatures
        self.logs: typing.Sequence[str] = logs

    @staticmethod
    def from_response(response: RPCResponse) -> "LogEvent":
        signature_text: str = response["result"]["value"]["signature"]
        signatures = signature_text.split(",")
        logs = response["result"]["value"]["logs"]
        return LogEvent(signatures, logs)

    def __str__(self) -> str:
        logs = "\n    ".join(self.logs)
        return f"""« LogEvent {self.signatures}
    {logs}
»"""

    def __repr__(self) -> str:
        return f"{self}"


class WebSocketLogSubscription(AddressWebSocketSubscription[LogEvent]):
    def __init__(self, context: Context, address: PublicKey) -> None:
        super().__init__(context, address, lambda _: LogEvent([""], []))

    def build_request(self) -> str:
        return (
            """
{
    "jsonrpc": "2.0",
    "id": \""""
            + str(self.id)
            + """\",
    "method": "logsSubscribe",
    "params": [
        {
            "mentions": [ \""""
            + str(self.address)
            + """\" ]
        },
        {
            "commitment": \""""
            + str(self.context.client.commitment)
            + """\"
        }
    ]
}
"""
        )

    def build(self, response: RPCResponse) -> LogEvent:
        return LogEvent.from_response(response)


# # 🥭 WebSocketSubscriptionManager class
#
# The `WebSocketSubscriptionManager` is a base class for different websocket management approaches.
#
class WebSocketSubscriptionManager(RxDisposable, metaclass=abc.ABCMeta):
    def __init__(self, context: Context, ping_interval: int = 10) -> None:
        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self.context: Context = context
        self.ping_interval: int = ping_interval
        self.subscriptions: typing.List[WebSocketSubscription[typing.Any]] = []

    def add(self, subscription: WebSocketSubscription[typing.Any]) -> None:
        self.subscriptions += [subscription]

    @abc.abstractmethod
    def open(self) -> None:
        raise NotImplementedError(
            "WebSocketSubscription.build_request() is not implemented on the base type."
        )

    @abc.abstractmethod
    def close(self) -> None:
        raise NotImplementedError(
            "WebSocketSubscription.build_request() is not implemented on the base type."
        )

    def dispose(self) -> None:
        for subscription in self.subscriptions:
            subscription.dispose()


@dataclass
class ActiveWebSocket:
    websocket: ReconnectingWebsocket
    subscription: WebSocketSubscription[typing.Any]
    pong_subscription: RxDisposable

    @staticmethod
    def create(
        context: Context,
        subscription: WebSocketSubscription[typing.Any],
        pong: BehaviorSubject,
    ) -> "ActiveWebSocket":
        def on_open(sock: websocket.WebSocketApp) -> None:
            sock.send(subscription.build_request())

        ws: ReconnectingWebsocket = ReconnectingWebsocket(
            context.client.cluster_ws_url, on_open
        )
        ws.item.subscribe(on_next=subscription._on_item)  # type: ignore[call-arg]
        ws.ping_interval = context.ping_interval
        ws.open()
        pong_subscription = ws.pong.subscribe(pong)

        return ActiveWebSocket(ws, subscription, pong_subscription)

    def dispose(self) -> None:
        self.pong_subscription.dispose()
        self.websocket.close()
        self.subscription.dispose()


# # 🥭 IndividualWebSocketSubscriptionManager class
#
# The `IndividualWebSocketSubscriptionManager` runs `WebSocketSubscription`s each in their
# own websocket.
#
class IndividualWebSocketSubscriptionManager(WebSocketSubscriptionManager):
    def __init__(self, context: Context, ping_interval: int = 10) -> None:
        super().__init__(context, ping_interval)
        self.pong: BehaviorSubject = BehaviorSubject(local_now())
        self.__subscriptions: typing.Sequence[ActiveWebSocket] = []

    def open(self) -> None:
        individual_subscriptions: typing.List[ActiveWebSocket] = []
        for subscription in self.subscriptions:
            individual_subscriptions += [
                ActiveWebSocket.create(self.context, subscription, self.pong)
            ]

        self.__subscriptions = individual_subscriptions

    def close(self) -> None:
        for subscription in self.__subscriptions:
            subscription.dispose()
        self.__subscriptions = []

    def dispose(self) -> None:
        super().dispose()
        self.close()


# # 🥭 SharedWebSocketSubscriptionManager class
#
# The `SharedWebSocketSubscriptionManager` runs a single websocket and sends updates to the correct
# `WebSocketSubscription`.
#
class SharedWebSocketSubscriptionManager(WebSocketSubscriptionManager):
    def __init__(self, context: Context, ping_interval: int = 10) -> None:
        super().__init__(context, ping_interval)
        self.ws: typing.Optional[ReconnectingWebsocket] = None
        self.pong: BehaviorSubject = BehaviorSubject(local_now())
        self._pong_subscription: typing.Optional[RxDisposable] = None

    def add(self, subscription: WebSocketSubscription[typing.Any]) -> None:
        self.subscriptions += [subscription]
        if self.ws is not None:
            request = subscription.build_request()
            self._logger.info(f"Sending request {request}")
            self.ws.send(subscription.build_request())

    def open(self) -> None:
        websocket_url = self.context.client.cluster_ws_url
        ws: ReconnectingWebsocket = ReconnectingWebsocket(
            websocket_url, self.open_handler
        )
        ws.item.subscribe(on_next=self.on_item)  # type: ignore[call-arg]
        ws.ping_interval = self.ping_interval
        self.ws = ws
        ws.open()
        self._pong_subscription = ws.pong.subscribe(self.pong)

    def close(self) -> None:
        if self.ws is not None:
            if self._pong_subscription is not None:
                self._pong_subscription.dispose()
                self._pong_subscription = None
            self.ws.close()
            self.ws = None

    def add_subscription_id(self, id: int, subscription_id: int) -> None:
        for subscription in self.subscriptions:
            if subscription.id == id:
                self._logger.info(
                    f"Setting ID {subscription_id} on subscription {subscription.id}."
                )
                subscription.subscription_id = subscription_id
                return
        self._logger.error(f"[{self.context.name}] Subscription ID {id} not found")

    def subscription_by_subscription_id(
        self, subscription_id: int
    ) -> WebSocketSubscription[typing.Any]:
        for subscription in self.subscriptions:
            if subscription.subscription_id == subscription_id:
                return subscription
        raise Exception(
            f"[{self.context.name}] No subscription with subscription ID {subscription_id} could be found."
        )

    def on_item(self, response: typing.Dict[str, typing.Any]) -> None:
        if "method" not in response:
            id: int = int(response["id"])
            subscription_id: int = int(response["result"])
            self.add_subscription_id(id, subscription_id)
        elif (
            (response["method"] == "accountNotification")
            or (response["method"] == "programNotification")
            or (response["method"] == "logsNotification")
            or (response["method"] == "signatureNotification")
        ):
            subscription_id = response["params"]["subscription"]
            subscription = self.subscription_by_subscription_id(subscription_id)
            built = subscription.build_subscribed_instance(response["params"])
            subscription.publisher.publish(built)
        else:
            self._logger.error(f"[{self.context.name}] Unknown response: {response}")

    def open_handler(self, ws: websocket.WebSocketApp) -> None:
        for subscription in self.subscriptions:
            ws.send(subscription.build_request())

    def dispose(self) -> None:
        super().dispose()
        if self.ws is not None:
            if self._pong_subscription is not None:
                self._pong_subscription.dispose()
                self._pong_subscription = None
            self.ws.close()
            self.ws = None



# # ⚠ Warning
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
# LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
# NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# [🥭 Entropy Markets](https://entropy.trade/) support is available at:
#   [Docs](https://docs.entropy.trade/)
#   [Discord](https://discord.gg/67jySBhxrg)
#   [Twitter](https://twitter.com/entropymarkets)
#   [Github](https://github.com/blockworks-foundation)
#   [Email](mailto:hello@blockworks.foundation)

import argparse
import entropy
import typing

from decimal import Decimal

from .element import Element
from ...modelstate import ModelState


# # 🥭 ConfidenceIntervalElement class
#
# Ignores any input `Order`s (so probably best at the head of the chain). Builds orders using a fixed position
# size ratio but with a spread based on the confidence in the oracle price.
#
class ConfidenceIntervalElement(Element):
    def __init__(
        self,
        order_type: entropy.OrderType,
        expire_seconds: typing.Optional[Decimal],
        match_limit: int,
        position_size_ratio: Decimal,
        confidence_interval_levels: typing.Sequence[Decimal],
    ) -> None:
        super().__init__()
        self.order_type: entropy.OrderType = order_type
        self.expire_seconds: typing.Optional[Decimal] = expire_seconds
        self.match_limit: int = match_limit
        self.position_size_ratio: Decimal = position_size_ratio
        self.confidence_interval_levels: typing.Sequence[
            Decimal
        ] = confidence_interval_levels

    @staticmethod
    def add_command_line_parameters(parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--confidenceinterval-level",
            type=Decimal,
            action="append",
            help="the levels of weighting to apply to the confidence interval from the oracle: e.g. 1 - use the oracle confidence interval as the spread, 2 (risk averse, default) - multiply the oracle confidence interval by 2 to get the spread, 0.5 (aggressive) halve the oracle confidence interval to get the spread (can be specified multiple times to give multiple levels)",
        )
        parser.add_argument(
            "--confidenceinterval-position-size-ratio",
            type=Decimal,
            help="fraction of the token inventory to be bought or sold in each order",
        )

    @staticmethod
    def from_command_line_parameters(
        args: argparse.Namespace,
    ) -> "ConfidenceIntervalElement":
        if (
            args.confidenceinterval_position_size_ratio is None
            or args.confidenceinterval_position_size_ratio == 0
        ):
            raise Exception("No position-size ratio specified.")

        order_type: entropy.OrderType = args.order_type
        expire_seconds: typing.Optional[Decimal] = args.expire_seconds
        match_limit: int = args.match_limit or entropy.Order.DefaultMatchLimit
        position_size_ratio: Decimal = args.confidenceinterval_position_size_ratio
        confidence_interval_levels: typing.Sequence[
            Decimal
        ] = args.confidenceinterval_level
        if len(confidence_interval_levels) == 0:
            confidence_interval_levels = [Decimal(2)]

        return ConfidenceIntervalElement(
            order_type,
            expire_seconds,
            match_limit,
            position_size_ratio,
            confidence_interval_levels,
        )

    def process(
        self,
        context: entropy.Context,
        model_state: ModelState,
        orders: typing.Sequence[entropy.Order],
    ) -> typing.Sequence[entropy.Order]:
        price: entropy.Price = model_state.price
        expiration = entropy.Order.build_absolute_expiration(self.expire_seconds)
        if price.source.supports.has_feature(entropy.SupportedOracleFeature.CONFIDENCE):
            raise Exception(f"Price does not support confidence interval: {price}")

        quote_value_to_risk = (
            model_state.inventory.available_collateral.value * self.position_size_ratio
        )
        position_size = quote_value_to_risk / price.mid_price

        new_orders: typing.List[entropy.Order] = []
        for confidence_interval_level in self.confidence_interval_levels:
            charge = price.confidence * confidence_interval_level
            bid: Decimal = price.mid_price - charge
            ask: Decimal = price.mid_price + charge

            bid_order = entropy.Order.from_values(
                entropy.Side.BUY,
                price=bid,
                quantity=position_size,
                order_type=self.order_type,
                expiration=expiration,
                match_limit=self.match_limit,
            )
            ask_order = entropy.Order.from_values(
                entropy.Side.SELL,
                price=ask,
                quantity=position_size,
                order_type=self.order_type,
                expiration=expiration,
                match_limit=self.match_limit,
            )
            self._logger.debug(
                f"""Desired orders:
    Bid: {bid_order}
    Ask: {ask_order}"""
            )
            new_orders += [bid_order, ask_order]

        new_orders.sort(key=lambda ord: ord.price, reverse=True)
        order_text = "\n    ".join([f"{order}" for order in new_orders])

        top_bid = model_state.top_bid
        top_ask = model_state.top_ask
        self._logger.debug(
            f"""Initial desired orders - spread {model_state.spread} ({top_bid.price if top_bid else None} / {top_ask.price if top_ask else None}):
    {order_text}"""
        )
        return new_orders

    def __str__(self) -> str:
        confidence_interval_levels = (
            ", ".join(map(str, self.confidence_interval_levels)) or "None"
        )
        return f"« ConfidenceIntervalElement {self.order_type} - position size: {self.position_size_ratio}, confidence interval levels: {confidence_interval_levels} »"
